"""IO handling within MQT Core."""

from __future__ import annotations

from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING

from . import QuantumComputation

if TYPE_CHECKING:
    from qiskit.circuit import QuantumCircuit


def load(input_circuit: QuantumComputation | str | PathLike[str] | QuantumCircuit) -> QuantumComputation:
    """Load a quantum circuit from any supported format as a ``QuantumComputation``.

    Args:
        input_circuit: The input circuit to translate to a ``QuantumComputation``.
        This can be a `QuantumComputation` itself, a file name to any of the supported file formats,
        an OpenQASM (2.0 or 3.0) string, or a Qiskit `QuantumCircuit`.

    Returns:
        The ``QuantumComputation``.

    Raises:
        ValueError: If the input circuit is a Qiskit `QuantumCircuit` but the `qiskit` extra is not installed.
        FileNotFoundError: If the input circuit is a file name and the file does not exist.
    """
    if isinstance(input_circuit, QuantumComputation):
        return input_circuit

    if isinstance(input_circuit, (str, PathLike)):
        if not Path(input_circuit).is_file():
            if isinstance(input_circuit, PathLike) or "OPENQASM" not in input_circuit:
                msg = f"File {input_circuit} does not exist."
                raise FileNotFoundError(msg)
            # otherwise, we assume that this is a QASM string
            return QuantumComputation.from_qasm(input_circuit)

        return QuantumComputation(input_circuit)

    try:
        from .plugins.qiskit import qiskit_to_mqt
    except ImportError:
        msg = "Qiskit is not installed. Please install `mqt.core[qiskit]` to use Qiskit circuits as input."
        raise ValueError(msg) from None

    return qiskit_to_mqt(input_circuit)


__all__ = ["load"]



from collections.abc import Iterator, Mapping, Sequence
from typing import overload

from .._compat.typing import Self

class Variable:
    def __eq__(self: Self, arg0: object) -> bool: ...
    def __gt__(self: Self, arg0: Variable) -> bool: ...
    def __hash__(self: Self) -> int: ...
    def __init__(self: Self, name: str = "") -> None: ...
    def __lt__(self: Self, arg0: Variable) -> bool: ...
    def __ne__(self: Self, arg0: object) -> bool: ...
    @property
    def name(self: Self) -> str: ...

class Term:
    def __eq__(self: Self, arg0: object) -> bool: ...
    def __hash__(self: Self) -> int: ...
    def __init__(self: Self, variable: Variable, coefficient: float = 1.0) -> None: ...
    def __mul__(self: Self, arg0: float) -> Term: ...
    def __ne__(self: Self, arg0: object) -> bool: ...
    def __rmul__(self: Self, arg0: float) -> Term: ...
    def __rtruediv__(self: Self, arg0: float) -> Term: ...
    def __truediv__(self: Self, arg0: float) -> Term: ...
    def add_coefficient(self: Self, coeff: float) -> None: ...
    def evaluate(self: Self, assignment: Mapping[Variable, float]) -> float: ...
    def has_zero_coefficient(self: Self) -> bool: ...
    @property
    def coefficient(self: Self) -> float: ...
    @property
    def variable(self: Self) -> Variable: ...

class Expression:
    constant: float
    @overload
    def __add__(self: Self, arg0: Expression) -> Expression: ...
    @overload
    def __add__(self: Self, arg0: Term) -> Expression: ...
    @overload
    def __add__(self: Self, arg0: float) -> Expression: ...
    def __eq__(self: Self, arg0: object) -> bool: ...
    def __getitem__(self: Self, idx: int) -> Term: ...
    def __hash__(self: Self) -> int: ...
    @overload
    def __init__(self: Self, terms: Sequence[Term], constant: float = 0.0) -> None: ...
    @overload
    def __init__(self: Self, term: Term, constant: float = 0.0) -> None: ...
    @overload
    def __init__(self: Self, constant: float = 0.0) -> None: ...
    def __iter__(self: Self) -> Iterator[Term]: ...
    def __len__(self: Self) -> int: ...
    def __mul__(self: Self, arg0: float) -> Expression: ...
    def __ne__(self: Self, arg0: object) -> bool: ...
    @overload
    def __radd__(self: Self, arg0: Term) -> Expression: ...
    @overload
    def __radd__(self: Self, arg0: float) -> Expression: ...
    def __rmul__(self: Self, arg0: float) -> Expression: ...
    @overload
    def __rsub__(self: Self, arg0: Term) -> Expression: ...
    @overload
    def __rsub__(self: Self, arg0: float) -> Expression: ...
    def __rtruediv__(self: Self, arg0: float) -> Expression: ...
    @overload
    def __sub__(self: Self, arg0: Expression) -> Expression: ...
    @overload
    def __sub__(self: Self, arg0: Term) -> Expression: ...
    @overload
    def __sub__(self: Self, arg0: float) -> Expression: ...
    def __truediv__(self: Self, arg0: float) -> Expression: ...
    def evaluate(self: Self, assignment: Mapping[Variable, float]) -> float: ...
    def is_constant(self: Self) -> bool: ...
    def is_zero(self: Self) -> bool: ...
    def num_terms(self: Self) -> int: ...
    @property
    def terms(self: Self) -> list[Term]: ...
    @property
    def variables(self: Self) -> set[Variable]: ...



"""Command line interface for mqt-core."""

from __future__ import annotations

import argparse
import sys

from ._version import version as __version__
from .commands import cmake_dir, include_dir


def main() -> None:
    """Entry point for the mqt-core command line interface."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=__version__, help="Print version and exit.")

    parser.add_argument(
        "--include_dir", action="store_true", help="Print the path to the mqt-core C++ include directory."
    )
    parser.add_argument(
        "--cmake_dir", action="store_true", help="Print the path to the mqt-core CMake module directory."
    )
    args = parser.parse_args()
    if not sys.argv[1:]:
        parser.print_help()
    if args.include_dir:
        print(include_dir().resolve())
    if args.cmake_dir:
        print(cmake_dir().resolve())


if __name__ == "__main__":
    main()



"""Operations Module."""

from __future__ import annotations

from ._core.operations import (
    ClassicControlledOperation,
    CompoundOperation,
    Control,
    NonUnitaryOperation,
    Operation,
    OpType,
    StandardOperation,
    SymbolicOperation,
)

__all__ = (
    "ClassicControlledOperation",
    "CompoundOperation",
    "Control",
    "NonUnitaryOperation",
    "OpType",
    "Operation",
    "StandardOperation",
    "SymbolicOperation",
)

for cls in (
    ClassicControlledOperation,
    CompoundOperation,
    Control,
    NonUnitaryOperation,
    Operation,
    OpType,
    StandardOperation,
    SymbolicOperation,
):
    cls.__module__ = "mqt.core.operations"
del cls



from abc import ABCMeta, abstractmethod
from collections.abc import Mapping, Sequence
from typing import ClassVar, overload

from .._compat.typing import Self
from ..symbolic import Expression, Variable

class Control:
    class Type:
        __members__: ClassVar[dict[Control.Type, str]]
        Neg: ClassVar[Control.Type]
        Pos: ClassVar[Control.Type]

        def __eq__(self: Self, other: object) -> bool: ...
        def __getstate__(self: Self) -> int: ...
        def __hash__(self: Self) -> int: ...
        def __index__(self: Self) -> int: ...
        def __init__(self: Self, value: int) -> None: ...
        def __int__(self: Self) -> int: ...
        def __ne__(self: Self, other: object) -> bool: ...
        def __setstate__(self: Self, state: int) -> None: ...
        @property
        def name(self: Self) -> str: ...
        @property
        def value(self: Self) -> int: ...

    qubit: int
    type_: Type

    @overload
    def __init__(self: Self, qubit: int) -> None: ...
    @overload
    def __init__(self: Self, qubit: int, type_: Type) -> None: ...

class OpType:
    __members__: ClassVar[dict[OpType, str]]  # readonly
    barrier: ClassVar[OpType]  # value = <OpType.barrier: 3>
    classic_controlled: ClassVar[OpType]  # value = <OpType.classic_controlled: 39>
    compound: ClassVar[OpType]  # value = <OpType.compound: 35>
    dcx: ClassVar[OpType]  # value = <OpType.dcx: 27>
    ecr: ClassVar[OpType]  # value = <OpType.ecr: 28>
    gphase: ClassVar[OpType]  # value = <OpType.gphase: 1>
    h: ClassVar[OpType]  # value = <OpType.h: 4>
    i: ClassVar[OpType]  # value = <OpType.i: 2>
    iswap: ClassVar[OpType]  # value = <OpType.iswap: 23>
    iswapdg: ClassVar[OpType]  # value = <OpType.iswapdg: 24>
    measure: ClassVar[OpType]  # value = <OpType.measure: 36>
    none: ClassVar[OpType]  # value = <OpType.none: 0>
    peres: ClassVar[OpType]  # value = <OpType.peres: 25>
    peresdg: ClassVar[OpType]  # value = <OpType.peresdg: 26>
    p: ClassVar[OpType]  # value = <OpType.p: 16>
    reset: ClassVar[OpType]  # value = <OpType.reset: 37>
    rx: ClassVar[OpType]  # value = <OpType.rx: 19>
    rxx: ClassVar[OpType]  # value = <OpType.rxx: 29>
    ry: ClassVar[OpType]  # value = <OpType.ry: 20>
    ryy: ClassVar[OpType]  # value = <OpType.ryy: 30>
    rz: ClassVar[OpType]  # value = <OpType.rz: 21>
    rzx: ClassVar[OpType]  # value = <OpType.rzx: 32>
    rzz: ClassVar[OpType]  # value = <OpType.rzz: 31>
    s: ClassVar[OpType]  # value = <OpType.s: 8>
    sdg: ClassVar[OpType]  # value = <OpType.sdg: 9>
    swap: ClassVar[OpType]  # value = <OpType.swap: 22>
    sx: ClassVar[OpType]  # value = <OpType.sx: 17>
    sxdg: ClassVar[OpType]  # value = <OpType.sxdg: 18>
    t: ClassVar[OpType]  # value = <OpType.t: 10>
    tdg: ClassVar[OpType]  # value = <OpType.tdg: 11>
    teleportation: ClassVar[OpType]  # value = <OpType.teleportation: 38>
    u2: ClassVar[OpType]  # value = <OpType.u2: 15>
    u: ClassVar[OpType]  # value = <OpType.u: 14>
    v: ClassVar[OpType]  # value = <OpType.v: 12>
    vdg: ClassVar[OpType]  # value = <OpType.vdg: 13>
    x: ClassVar[OpType]  # value = <OpType.x: 5>
    xx_minus_yy: ClassVar[OpType]  # value = <OpType.xx_minus_yy: 33>
    xx_plus_yy: ClassVar[OpType]  # value = <OpType.xx_plus_yy: 34>
    y: ClassVar[OpType]  # value = <OpType.y: 6>
    z: ClassVar[OpType]  # value = <OpType.z: 7>
    @property
    def name(self: Self) -> str: ...
    @property
    def value(self: Self) -> int: ...
    def __eq__(self: Self, other: object) -> bool: ...
    def __getstate__(self: Self) -> int: ...
    def __hash__(self: Self) -> int: ...
    def __index__(self: Self) -> int: ...
    @overload
    def __init__(self: Self, value: int) -> None: ...
    @overload
    def __init__(self: Self, arg0: str) -> None: ...
    def __int__(self: Self) -> int: ...
    def __ne__(self: Self, other: object) -> bool: ...
    def __setstate__(self: Self, state: int) -> None: ...

class Operation(metaclass=ABCMeta):
    type_: OpType
    controls: set[Control]
    num_qubits: int
    targets: list[int]
    parameter: list[float]
    @property
    def name(self: Self) -> str: ...
    @property
    def num_targets(self: Self) -> int: ...
    @property
    def num_controls(self: Self) -> int: ...
    @abstractmethod
    def add_control(self: Self, control: Control) -> None: ...
    def add_controls(self: Self, controls: set[Control]) -> None: ...
    @abstractmethod
    def clear_controls(self: Self) -> None: ...
    @abstractmethod
    def remove_control(self: Self, control: Control) -> None: ...
    def remove_controls(self: Self, controls: set[Control]) -> None: ...
    def acts_on(self: Self, qubit: int) -> bool: ...
    def get_used_qubits(self: Self) -> set[int]: ...
    def is_classic_controlled_operation(self: Self) -> bool: ...
    def is_compound_operation(self: Self) -> bool: ...
    def is_controlled(self: Self) -> bool: ...
    def is_non_unitary_operation(self: Self) -> bool: ...
    def is_standard_operation(self: Self) -> bool: ...
    def is_symbolic_operation(self: Self) -> bool: ...
    def is_unitary(self: Self) -> bool: ...
    def get_inverted(self: Self) -> Operation: ...
    @abstractmethod
    def invert(self: Self) -> None: ...
    @abstractmethod
    def qasm2_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...
    @abstractmethod
    def qasm3_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...
    def __eq__(self: Self, other: object) -> bool: ...
    def __hash__(self: Self) -> int: ...
    def __ne__(self: Self, other: object) -> bool: ...

class StandardOperation(Operation):
    @overload
    def __init__(self: Self) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        target: int,
        op_type: OpType,
        params: Sequence[float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        targets: Sequence[int],
        op_type: OpType,
        params: Sequence[float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        control: Control,
        target: int,
        op_type: OpType,
        params: Sequence[float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        control: Control,
        targets: Sequence[int],
        op_type: OpType,
        params: Sequence[float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        controls: set[Control],
        target: int,
        op_type: OpType,
        params: Sequence[float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        controls: set[Control],
        targets: Sequence[int],
        op_type: OpType,
        params: Sequence[float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(self: Self, nq: int, controls: set[Control], target: int, starting_qubit: int = 0) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        controls: set[Control],
        target0: int,
        target1: int,
        op_type: OpType,
        params: Sequence[float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    def add_control(self: Self, control: Control) -> None: ...
    def clear_controls(self: Self) -> None: ...
    def remove_control(self: Self, control: Control) -> None: ...
    def invert(self: Self) -> None: ...
    def qasm2_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...
    def qasm3_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...

class NonUnitaryOperation(Operation):
    @property
    def classics(self: Self) -> list[int]: ...
    @overload
    def __init__(self: Self, nq: int, targets: Sequence[int], classics: Sequence[int]) -> None: ...
    @overload
    def __init__(self: Self, nq: int, target: int, classic: int) -> None: ...
    @overload
    def __init__(self: Self, nq: int, targets: Sequence[int], op_type: OpType = ...) -> None: ...
    def add_control(self: Self, control: Control) -> None: ...
    def clear_controls(self: Self) -> None: ...
    def remove_control(self: Self, control: Control) -> None: ...
    def invert(self: Self) -> None: ...
    def qasm2_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...
    def qasm3_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...

class CompoundOperation(Operation):
    @overload
    def __init__(self: Self, nq: int) -> None: ...
    @overload
    def __init__(self: Self, nq: int, ops: Sequence[Operation]) -> None: ...
    def __len__(self: Self) -> int: ...
    @overload
    def __getitem__(self: Self, idx: int) -> Operation: ...
    @overload
    def __getitem__(self: Self, idx: slice) -> list[Operation]: ...
    def append(self: Self, op: Operation) -> None: ...
    def empty(self: Self) -> bool: ...
    def add_control(self: Self, control: Control) -> None: ...
    def clear_controls(self: Self) -> None: ...
    def remove_control(self: Self, control: Control) -> None: ...
    def invert(self: Self) -> None: ...
    def qasm2_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...
    def qasm3_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...

class SymbolicOperation(StandardOperation):
    @overload
    def __init__(self: Self) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        target: int,
        op_type: OpType,
        params: Sequence[Expression | float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        targets: Sequence[int],
        op_type: OpType,
        params: Sequence[Expression | float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        control: Control,
        target: int,
        op_type: OpType,
        params: Sequence[Expression | float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        control: Control,
        targets: Sequence[int],
        op_type: OpType,
        params: Sequence[Expression | float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        controls: set[Control],
        target: int,
        op_type: OpType,
        params: Sequence[Expression | float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        controls: set[Control],
        targets: Sequence[int],
        op_type: OpType,
        params: Sequence[Expression | float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        nq: int,
        controls: set[Control],
        target0: int,
        target1: int,
        op_type: OpType,
        params: Sequence[Expression | float] | None = None,
        starting_qubit: int = 0,
    ) -> None: ...
    def get_parameter(self: Self, idx: int) -> Expression | float: ...
    def get_parameters(self: Self) -> list[Expression | float]: ...
    def get_instantiated_operation(self: Self, assignment: Mapping[Variable, float]) -> StandardOperation: ...
    def instantiate(self: Self, assignment: Mapping[Variable, float]) -> None: ...

class ClassicControlledOperation(Operation):
    def __init__(
        self: Self, operation: Operation, control_register: tuple[int, int], expected_value: int = 1
    ) -> None: ...
    @property
    def operation(self: Self) -> Operation: ...
    @property
    def control_register(self: Self) -> tuple[int, int]: ...
    @property
    def expected_value(self: Self) -> int: ...
    def add_control(self: Self, control: Control) -> None: ...
    def clear_controls(self: Self) -> None: ...
    def remove_control(self: Self, control: Control) -> None: ...
    def invert(self: Self) -> None: ...
    def qasm2_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...
    def qasm3_str(self: Self, qreg: Sequence[tuple[str, str]], creg: Sequence[tuple[str, str]]) -> str: ...



"""MQT Core - The Backbone of the Munich Quantum Toolkit."""

from __future__ import annotations

from ._core import Permutation, QuantumComputation
from ._version import version as __version__
from ._version import version_tuple as version_info
from .commands import cmake_dir, include_dir

__all__ = [
    "Permutation",
    "QuantumComputation",
    "__version__",
    "cmake_dir",
    "include_dir",
    "version_info",
]

for cls in (Permutation, QuantumComputation):
    cls.__module__ = "mqt.core"
del cls



from collections.abc import Iterable, Iterator, Mapping, MutableMapping, MutableSequence, Sequence
from os import PathLike
from typing import overload

from .._compat.typing import Self
from .operations import Control, Operation, OpType
from .symbolic import Expression, Variable

class Permutation(MutableMapping[int, int]):
    def __getitem__(self: Self, idx: int) -> int: ...
    def __setitem__(self: Self, idx: int, val: int) -> None: ...
    def __delitem__(self: Self, key: int) -> None: ...
    def __iter__(self: Self) -> Iterator[int]: ...
    def __len__(self: Self) -> int: ...
    @overload
    def apply(self: Self, controls: set[Control]) -> set[Control]: ...
    @overload
    def apply(self: Self, targets: list[int]) -> list[int]: ...

class QuantumComputation(MutableSequence[Operation]):
    # --------------------------------------------------------------------------
    #                               Constructors
    # --------------------------------------------------------------------------

    @overload
    def __init__(self: Self) -> None: ...
    @overload
    def __init__(self: Self, nq: int, nc: int = 0) -> None: ...
    @overload
    def __init__(self: Self, filename: str | PathLike[str]) -> None: ...
    @staticmethod
    def from_qasm(qasm: str) -> QuantumComputation: ...

    # --------------------------------------------------------------------------
    #                          General Properties
    # --------------------------------------------------------------------------

    name: str
    global_phase: float

    @property
    def num_qubits(self: Self) -> int: ...
    @property
    def num_ancilla_qubits(self: Self) -> int: ...
    @property
    def num_qubits_without_ancilla_qubits(self: Self) -> int: ...
    @property
    def num_classical_bits(self: Self) -> int: ...
    @property
    def num_ops(self: Self) -> int: ...
    def num_single_qubit_ops(self: Self) -> int: ...
    @property
    def num_total_ops(self: Self) -> int: ...
    def depth(self: Self) -> int: ...
    def invert(self: Self) -> None: ...
    def to_operation(self: Self) -> Operation: ...

    # --------------------------------------------------------------------------
    #                 Mutable Sequence Interface
    # --------------------------------------------------------------------------

    def __len__(self: Self) -> int: ...
    @overload
    def __getitem__(self: Self, idx: int) -> Operation: ...
    @overload
    def __getitem__(self: Self, idx: slice) -> list[Operation]: ...
    @overload
    def __setitem__(self: Self, idx: int, op: Operation) -> None: ...
    @overload
    def __setitem__(self: Self, idx: slice, ops: Iterable[Operation]) -> None: ...
    @overload
    def __delitem__(self: Self, idx: int) -> None: ...
    @overload
    def __delitem__(self: Self, idx: slice) -> None: ...
    def insert(self: Self, idx: int, op: Operation) -> None: ...
    def append(self: Self, op: Operation) -> None: ...
    def reverse(self: Self) -> None: ...
    def clear(self: Self) -> None: ...

    # --------------------------------------------------------------------------
    #                          (Qu)Bit Registers
    # --------------------------------------------------------------------------

    def add_ancillary_register(self: Self, n: int, name: str = "q") -> None: ...
    def add_classical_register(self: Self, n: int, name: str = "c") -> None: ...
    def add_qubit_register(self: Self, n: int, name: str = "anc") -> None: ...
    def unify_quantum_registers(self: Self, name: str = "q") -> None: ...

    # --------------------------------------------------------------------------
    #                  Initial Layout and Output Permutation
    # --------------------------------------------------------------------------

    initial_layout: Permutation
    output_permutation: Permutation

    def initialize_io_mapping(self: Self) -> None: ...

    # --------------------------------------------------------------------------
    #                       Ancilla and Garbage Handling
    # --------------------------------------------------------------------------

    @property
    def ancillary(self: Self) -> list[bool]: ...
    def set_circuit_qubit_ancillary(self: Self, q: int) -> None: ...
    def is_circuit_qubit_ancillary(self: Self, q: int) -> bool: ...
    @property
    def garbage(self: Self) -> list[bool]: ...
    def set_circuit_qubit_garbage(self: Self, q: int) -> None: ...
    def is_circuit_qubit_garbage(self: Self, q: int) -> bool: ...

    # --------------------------------------------------------------------------
    #                        Symbolic Circuit Handling
    # --------------------------------------------------------------------------

    @property
    def variables(self: Self) -> set[Variable]: ...
    def add_variable(self: Self, var: Expression | float) -> None: ...
    def add_variables(self: Self, vars_: Sequence[Expression | float]) -> None: ...
    def is_variable_free(self: Self) -> bool: ...
    def instantiate(self: Self, assignment: Mapping[Variable, float]) -> QuantumComputation: ...
    def instantiate_inplace(self: Self, assignment: Mapping[Variable, float]) -> None: ...

    # --------------------------------------------------------------------------
    #                             Output Handling
    # --------------------------------------------------------------------------

    def qasm2_str(self: Self) -> str: ...
    def qasm2(self: Self, filename: PathLike[str] | str) -> None: ...
    def qasm3_str(self: Self) -> str: ...
    def qasm3(self: Self, filename: PathLike[str] | str) -> None: ...

    # --------------------------------------------------------------------------
    #                               Operations
    # --------------------------------------------------------------------------

    def i(self: Self, q: int) -> None: ...
    def ci(self: Self, control: Control | int, target: int) -> None: ...
    def mci(self: Self, controls: set[Control | int], target: int) -> None: ...
    def x(self: Self, q: int) -> None: ...
    def cx(self: Self, control: Control | int, target: int) -> None: ...
    def mcx(self: Self, controls: set[Control | int], target: int) -> None: ...
    def y(self: Self, q: int) -> None: ...
    def cy(self: Self, control: Control | int, target: int) -> None: ...
    def mcy(self: Self, controls: set[Control | int], target: int) -> None: ...
    def z(self: Self, q: int) -> None: ...
    def cz(self: Self, control: Control | int, target: int) -> None: ...
    def mcz(self: Self, controls: set[Control | int], target: int) -> None: ...
    def h(self: Self, q: int) -> None: ...
    def ch(self: Self, control: Control | int, target: int) -> None: ...
    def mch(self: Self, controls: set[Control | int], target: int) -> None: ...
    def s(self: Self, q: int) -> None: ...
    def cs(self: Self, control: Control | int, target: int) -> None: ...
    def mcs(self: Self, controls: set[Control | int], target: int) -> None: ...
    def sdg(self: Self, q: int) -> None: ...
    def csdg(self: Self, control: Control | int, target: int) -> None: ...
    def mcsdg(self: Self, controls: set[Control | int], target: int) -> None: ...
    def t(self: Self, q: int) -> None: ...
    def ct(self: Self, control: Control | int, target: int) -> None: ...
    def mct(self: Self, controls: set[Control | int], target: int) -> None: ...
    def tdg(self: Self, q: int) -> None: ...
    def ctdg(self: Self, control: Control | int, target: int) -> None: ...
    def mctdg(self: Self, controls: set[Control | int], target: int) -> None: ...
    def v(self: Self, q: int) -> None: ...
    def cv(self: Self, control: Control | int, target: int) -> None: ...
    def mcv(self: Self, controls: set[Control | int], target: int) -> None: ...
    def vdg(self: Self, q: int) -> None: ...
    def cvdg(self: Self, control: Control | int, target: int) -> None: ...
    def mcvdg(self: Self, controls: set[Control | int], target: int) -> None: ...
    def sx(self: Self, q: int) -> None: ...
    def csx(self: Self, control: Control | int, target: int) -> None: ...
    def mcsx(self: Self, controls: set[Control | int], target: int) -> None: ...
    def sxdg(self: Self, q: int) -> None: ...
    def csxdg(self: Self, control: Control | int, target: int) -> None: ...
    def mcsxdg(self: Self, controls: set[Control | int], target: int) -> None: ...
    def rx(self: Self, theta: float | Expression, q: int) -> None: ...
    def crx(self: Self, theta: float | Expression, control: Control | int, target: int) -> None: ...
    def mcrx(self: Self, theta: float | Expression, controls: set[Control | int], target: int) -> None: ...
    def ry(self: Self, theta: float | Expression, q: int) -> None: ...
    def cry(self: Self, theta: float | Expression, control: Control | int, target: int) -> None: ...
    def mcry(self: Self, theta: float | Expression, controls: set[Control | int], target: int) -> None: ...
    def rz(self: Self, theta: float | Expression, q: int) -> None: ...
    def crz(self: Self, theta: float | Expression, control: Control | int, target: int) -> None: ...
    def mcrz(self: Self, theta: float | Expression, controls: set[Control | int], target: int) -> None: ...
    def p(self: Self, theta: float | Expression, q: int) -> None: ...
    def cp(self: Self, theta: float | Expression, control: Control | int, target: int) -> None: ...
    def mcp(self: Self, theta: float | Expression, controls: set[Control | int], target: int) -> None: ...
    def u2(self: Self, phi: float | Expression, lambda_: float | Expression, q: int) -> None: ...
    def cu2(
        self: Self, phi: float | Expression, lambda_: float | Expression, control: Control | int, target: int
    ) -> None: ...
    def mcu2(
        self: Self, phi: float | Expression, lambda_: float | Expression, controls: set[Control | int], target: int
    ) -> None: ...
    def u(
        self: Self, theta: float | Expression, phi: float | Expression, lambda_: float | Expression, q: int
    ) -> None: ...
    def cu(
        self: Self,
        theta: float | Expression,
        phi: float | Expression,
        lambda_: float | Expression,
        control: Control | int,
        target: int,
    ) -> None: ...
    def mcu(
        self: Self,
        theta: float | Expression,
        phi: float | Expression,
        lambda_: float | Expression,
        controls: set[Control | int],
        target: int,
    ) -> None: ...
    def swap(self: Self, target1: int, target2: int) -> None: ...
    def cswap(self: Self, control: Control | int, target1: int, target2: int) -> None: ...
    def mcswap(self: Self, controls: set[Control | int], target1: int, target2: int) -> None: ...
    def dcx(self: Self, target1: int, target2: int) -> None: ...
    def cdcx(self: Self, control: Control | int, target1: int, target2: int) -> None: ...
    def mcdcx(self: Self, controls: set[Control | int], target1: int, target2: int) -> None: ...
    def ecr(self: Self, target1: int, target2: int) -> None: ...
    def cecr(self: Self, control: Control | int, target1: int, target2: int) -> None: ...
    def mcecr(self: Self, controls: set[Control | int], target1: int, target2: int) -> None: ...
    def iswap(self: Self, target1: int, target2: int) -> None: ...
    def ciswap(self: Self, control: Control | int, target1: int, target2: int) -> None: ...
    def mciswap(self: Self, controls: set[Control | int], target1: int, target2: int) -> None: ...
    def iswapdg(self: Self, target1: int, target2: int) -> None: ...
    def ciswapdg(self: Self, control: Control | int, target1: int, target2: int) -> None: ...
    def mciswapdg(self: Self, controls: set[Control | int], target1: int, target2: int) -> None: ...
    def peres(self: Self, target1: int, target2: int) -> None: ...
    def cperes(self: Self, control: Control | int, target1: int, target2: int) -> None: ...
    def mcperes(self: Self, controls: set[Control | int], target1: int, target2: int) -> None: ...
    def peresdg(self: Self, target1: int, target2: int) -> None: ...
    def cperesdg(self: Self, control: Control | int, target1: int, target2: int) -> None: ...
    def mcperesdg(self: Self, controls: set[Control | int], target1: int, target2: int) -> None: ...
    def rxx(self: Self, theta: float | Expression, target1: int, target2: int) -> None: ...
    def crxx(self: Self, theta: float | Expression, control: Control | int, target1: int, target2: int) -> None: ...
    def mcrxx(
        self: Self, theta: float | Expression, controls: set[Control | int], target1: int, target2: int
    ) -> None: ...
    def ryy(self: Self, theta: float | Expression, target1: int, target2: int) -> None: ...
    def cryy(self: Self, theta: float | Expression, control: Control | int, target1: int, target2: int) -> None: ...
    def mcryy(
        self: Self, theta: float | Expression, controls: set[Control | int], target1: int, target2: int
    ) -> None: ...
    def rzz(self: Self, theta: float | Expression, target1: int, target2: int) -> None: ...
    def crzz(self: Self, theta: float | Expression, control: Control | int, target1: int, target2: int) -> None: ...
    def mcrzz(
        self: Self, theta: float | Expression, controls: set[Control | int], target1: int, target2: int
    ) -> None: ...
    def rzx(self: Self, theta: float | Expression, target1: int, target2: int) -> None: ...
    def crzx(self: Self, theta: float | Expression, control: Control | int, target1: int, target2: int) -> None: ...
    def mcrzx(
        self: Self, theta: float | Expression, controls: set[Control | int], target1: int, target2: int
    ) -> None: ...
    def xx_minus_yy(
        self: Self, theta: float | Expression, beta: float | Expression, target1: int, target2: int
    ) -> None: ...
    def cxx_minus_yy(
        self: Self,
        theta: float | Expression,
        beta: float | Expression,
        control: Control | int,
        target1: int,
        target2: int,
    ) -> None: ...
    def mcxx_minus_yy(
        self: Self,
        theta: float | Expression,
        beta: float | Expression,
        controls: set[Control | int],
        target1: int,
        target2: int,
    ) -> None: ...
    def xx_plus_yy(
        self: Self, theta: float | Expression, beta: float | Expression, target1: int, target2: int
    ) -> None: ...
    def cxx_plus_yy(
        self: Self,
        theta: float | Expression,
        beta: float | Expression,
        control: Control | int,
        target1: int,
        target2: int,
    ) -> None: ...
    def mcxx_plus_yy(
        self: Self,
        theta: float | Expression,
        beta: float | Expression,
        controls: set[Control | int],
        target1: int,
        target2: int,
    ) -> None: ...
    def gphase(self: Self, theta: float) -> None: ...
    @overload
    def measure(self: Self, qubit: int, cbit: int) -> None: ...
    @overload
    def measure(self: Self, qubit: int, creg_bit: tuple[str, int]) -> None: ...
    @overload
    def measure(self: Self, qubits: Sequence[int], cbits: Sequence[int]) -> None: ...
    def measure_all(self: Self, add_bits: bool = True) -> None: ...
    @overload
    def reset(self: Self, q: int) -> None: ...
    @overload
    def reset(self: Self, qubits: Sequence[int]) -> None: ...
    @overload
    def barrier(self: Self) -> None: ...
    @overload
    def barrier(self: Self, q: int) -> None: ...
    @overload
    def barrier(self: Self, qubits: Sequence[int]) -> None: ...
    @overload
    def classic_controlled(
        self: Self,
        op: OpType,
        target: int,
        creg: tuple[int, int],
        expected_value: int,
        params: Sequence[float],
    ) -> None: ...
    @overload
    def classic_controlled(
        self: Self,
        op: OpType,
        target: int,
        control: Control | int,
        creg: tuple[int, int],
        expected_value: int,
        params: Sequence[float],
    ) -> None: ...
    @overload
    def classic_controlled(
        self: Self,
        op: OpType,
        target: int,
        controls: set[Control | int],
        creg: tuple[int, int],
        expected_value: int,
        params: Sequence[float],
    ) -> None: ...

__all__ = [
    "Permutation",
    "QuantumComputation",
]



"""Symbolic expressions and terms."""

from __future__ import annotations

from ._core.symbolic import Expression, Term, Variable

__all__ = ("Expression", "Term", "Variable")

for cls in (Variable, Term, Expression):
    cls.__module__ = "mqt.core.symbolic"
del cls



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if sys.version_info < (3, 11):
    if TYPE_CHECKING:
        from typing_extensions import Self, assert_never
    else:
        Self = object

        def assert_never(_: object) -> None:
            msg = "Expected code to be unreachable"
            raise AssertionError(msg)

else:
    from typing import Self, assert_never

__all__ = ["Self", "assert_never"]


def __dir__() -> list[str]:
    return __all__
